;; Generated package description from colorful-mode.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "colorful-mode" "1.2.4" "Preview any color in your buffer in real time" '((emacs "28.1") (compat "30.1.0.0")) :commit "25ad453e4d4e6d9d9e9adf6ccc510231a6059483" :authors '(("Elias G. Perez" . "eg642616@gmail.com")) :maintainer '(("Jen-Chieh" . "jcs090218@gmail.com") ("Elias G. Perez" . "eg642616@gmail.com")) :keywords '("faces" "tools" "matching" "convenience") :url "https://github.com/DevelopmentCool2449/colorful-mode")
